@@ -1,4 +1,3 @@
-/* $XFree86: xc/programs/Xserver/mfb/mfbgc.c,v 1.9 2003/07/16 01:38:55 dawes Exp $ */
 /***********************************************************
 
 Copyright 1987, 1998  The Open Group
@@ -45,7 +44,6 @@ ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 SOFTWARE.
 
 ******************************************************************/
-/* $Xorg: mfbgc.c,v 1.4 2001/02/09 02:05:19 xorgcvs Exp $ */
 #ifdef HAVE_DIX_CONFIG_H
 #include <dix-config.h>
 #endif
@@ -101,9 +99,6 @@ static GCOps	whiteTECopyOps = {
 	mfbTEGlyphBltWhite,
 	mfbPolyGlyphBltWhite,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	blackTECopyOps = {
@@ -127,9 +122,6 @@ static GCOps	blackTECopyOps = {
 	mfbTEGlyphBltBlack,
 	mfbPolyGlyphBltBlack,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	whiteTEInvertOps = {
@@ -153,9 +145,6 @@ static GCOps	whiteTEInvertOps = {
 	mfbTEGlyphBltWhite,
 	mfbPolyGlyphBltInvert,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	blackTEInvertOps = {
@@ -179,9 +168,6 @@ static GCOps	blackTEInvertOps = {
 	mfbTEGlyphBltBlack,
 	mfbPolyGlyphBltInvert,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	whiteCopyOps = {
@@ -205,9 +191,6 @@ static GCOps	whiteCopyOps = {
 	mfbImageGlyphBltWhite,
 	mfbPolyGlyphBltWhite,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	blackCopyOps = {
@@ -231,9 +214,6 @@ static GCOps	blackCopyOps = {
 	mfbImageGlyphBltBlack,
 	mfbPolyGlyphBltBlack,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	whiteInvertOps = {
@@ -257,9 +237,6 @@ static GCOps	whiteInvertOps = {
 	mfbImageGlyphBltWhite,
 	mfbPolyGlyphBltInvert,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	blackInvertOps = {
@@ -283,9 +260,6 @@ static GCOps	blackInvertOps = {
 	mfbImageGlyphBltBlack,
 	mfbPolyGlyphBltInvert,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	whiteWhiteCopyOps = {
@@ -309,9 +283,6 @@ static GCOps	whiteWhiteCopyOps = {
 	miImageGlyphBlt,
 	mfbPolyGlyphBltWhite,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	blackBlackCopyOps = {
@@ -335,9 +306,6 @@ static GCOps	blackBlackCopyOps = {
 	miImageGlyphBlt,
 	mfbPolyGlyphBltBlack,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 static GCOps	fgEqBgInvertOps = {
@@ -361,9 +329,6 @@ static GCOps	fgEqBgInvertOps = {
 	miImageGlyphBlt,
 	mfbPolyGlyphBltInvert,
 	mfbSolidPP
-#ifdef NEED_LINEHELPER
-	,NULL
-#endif
 };
 
 
@@ -416,7 +381,8 @@ matchCommon (
 	FONTMINBOUNDS(pGC->font,leftSideBearing) > 32 ||
 	FONTMINBOUNDS(pGC->font,characterWidth) < 0)
 	return 0;
-    priv = (mfbPrivGC *) pGC->devPrivates[mfbGCPrivateIndex].ptr;
+    priv = (mfbPrivGC *)dixLookupPrivate(&pGC->devPrivates,
+					 mfbGetGCPrivateKey());
     for (i = 0; i < numberCommonOps; i++) {
 	cop = &mfbCommonOps[i];
 	if ((pGC->fgPixel & 1) != cop->fg)
@@ -455,7 +421,8 @@ mfbCreateGC(pGC)
     /* mfb wants to translate before scan convesion */
     pGC->miTranslate = 1;
 
-    pPriv = (mfbPrivGC *)(pGC->devPrivates[mfbGCPrivateIndex].ptr);
+    pPriv = (mfbPrivGC *)dixLookupPrivate(&pGC->devPrivates,
+					  mfbGetGCPrivateKey());
     pPriv->rop = mfbReduceRop(pGC->alu, pGC->fgPixel);
     pGC->fExpose = TRUE;
     pGC->pRotatedPixmap = NullPixmap;
@@ -543,8 +510,8 @@ mfbValidateGC(pGC, changes, pDrawable)
     new_rotate = (oldOrg.x != pGC->lastWinOrg.x) ||
 		 (oldOrg.y != pGC->lastWinOrg.y);
 
-    devPriv = ((mfbPrivGCPtr) (pGC->devPrivates[mfbGCPrivateIndex].ptr));
-
+    devPriv = (mfbPrivGCPtr)dixLookupPrivate(&pGC->devPrivates,
+					     mfbGetGCPrivateKey());
     /*
 	if the client clip is different or moved OR
 	the subwindowMode has changed OR
